////////////////////////////////////////////////////////////////////////////////
// File         : Baustein.cpp
// Subsystem    : Graph
// Component    : object networks
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : implementation of the class Baustein
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 01.10.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
// 30.05.06    Josef Huebl       some extensions
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
//     
//     
////////////////////////////////////////////////////////////////////////////////

#include "Baustein.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Baustein::Baustein( IsoTypeNode &class_isoType, int numberOfEntries, int numberOfExits, Baustein *owner, HnetUniverse *universe )
	: m_universe( (universe != NULL) ? universe : owner->m_universe ),
	m_classIsoTypePtr( &class_isoType ),
	m_numberOfEntries( numberOfEntries ), m_numberOfExits( numberOfExits ), m_numberOfSubbausteins( 0 ), m_elementary( false )
{
	m_isoType = CheckClassIsoType( class_isoType );

	if ( owner != NULL ) 
	{
		if ( !*(owner->m_classIsoTypePtr) && owner->m_isoType ) 
		{
			Hnet &hnetOwner = m_universe->GetHnet( owner->m_isoType );
			hnetOwner.AddSubcentre( m_isoType );
		}

		owner->AddSubbaustein( this );
	}

	// Attention! Because of STL it is not possible to store m_subcentre here
}


Baustein::~Baustein()
{

}


//////////////////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////////////////

// Attention! Outer entries are inner exits and outer exits are inner entries

void Baustein::Connect( Baustein &baustein1, int exitPos, Baustein &baustein2, int entryPos )
{ 
	if (!CheckConnect( baustein1, exitPos, baustein2, entryPos ))
		return;
		
	Connect( baustein1.Exit( exitPos ), baustein2.Entry( entryPos ) ); 
}


void Baustein::ConnectParallel( Baustein &baustein1, int exitPos, Baustein &baustein2, int entryPos, int numberOfConnections )
{ 
	if (!CheckConnect( baustein1, exitPos, baustein2, entryPos, numberOfConnections, numberOfConnections ))
		return;

	for ( int i = 0; i < numberOfConnections; i++ )
		Connect( baustein1.Exit( exitPos+i ), baustein2.Entry( entryPos+i ) ); 
}


void Baustein::ConnectAll( Baustein &baustein1, int exitPos, Baustein &baustein2, int entryPos, int numberOfExits, int numberOfEntries )
{ 
	if (!CheckConnect( baustein1, exitPos, baustein2, entryPos, numberOfExits, numberOfEntries ))
		return;

	for ( int i = 0; i < numberOfExits; i++ )
		for ( int j = 0; j < numberOfEntries; j++ )
			Connect( baustein1.Exit( exitPos+i ), baustein2.Entry( entryPos+j ) ); 
}


void Baustein::SetSubcentre( const Node &s )
{ 
	m_subcentre = s; 

	m_entry.DelAll(); // because these are used before by Root()
	m_exit.DelAll();

	Hnet &hnet = OwnHnet();

	// get all entries and exits as subcenter in hnetOwner
	Edge e;
	Node n = hnet.FirstEntry( e, m_subcentre );

	while ( n )
	{
		m_entry.AddLast( n );
		n = hnet.NextEntry( e );
	}

	n = hnet.FirstExit( e, m_subcentre );

	while ( n )
	{
		m_exit.AddLast( n );
		n = hnet.NextExit( e );
	}

}



void Baustein::MapSubcentresElementary( Hnet &hnet, const Node &s )
{
	int i;
	Node ss, sFirst;
	Node sNext;

	ss = s;

	for( i = 0; i <	m_numberOfSubbausteins; i++)
	{
		if ( m_subbausteins[ i ]->m_elementary )
		{
			m_subbausteins[ i ]->SetSubcentre( ss );
			ss = hnet.NextSubcentre( ss );
		}
		else
		{
			sNext = hnet.NextSubcentre( ss ); 
			sFirst = hnet.Expand( ss ); // ss is deleted
			m_subbausteins[ i ]->MapSubcentresElementary( hnet, sFirst ); 
			ss = sNext;
		}
	}
}


//////////////////////////////////////////////////////////////////////
// protected
//////////////////////////////////////////////////////////////////////

void Baustein::BuiltClassIsoType( IsoTypeNode &class_isoType )
{
	Hnet &hnet = OwnHnet();

	SetSubcentre( hnet.Root() ); //store outer entries and exits temporarly, until connections are created

	// store subcentres with its entries and exits temporarly, until connections are created
	Node s = hnet.FirstSubcentre();

	for ( int i=0; i<m_numberOfSubbausteins; i++ )
	{
		m_subbausteins[i]->SetSubcentre( s ); 
		s = hnet.NextSubcentre( s );
	}

	if ( !class_isoType )
	{
		MakeConnections(); // imlemented by the derived class
		class_isoType = m_isoType;
	}
}


//////////////////////////////////////////////////////////////////////
// private
//////////////////////////////////////////////////////////////////////

IsoTypeNode Baustein::CheckClassIsoType( IsoTypeNode &class_isoType )
{
	if ( class_isoType ) // does already one instance of Baustein exist?
		 return class_isoType;
	else
		return m_universe->CreateIsoType( m_numberOfEntries, m_numberOfExits );
}



bool Baustein::CheckConnect( Baustein &baustein1, int exitPos, Baustein &baustein2, int entryPos, int numberOfExits, int numberOfEntries )
{
	int noExits, noEntries;

	if( &baustein1 == this )
		noExits = baustein1.m_numberOfEntries;
	else
		noExits = baustein1.m_numberOfExits;

	if( &baustein2 == this )
		noEntries = baustein2.m_numberOfExits;
	else
		noEntries = baustein2.m_numberOfEntries;

	if (( exitPos >= noExits ) || (exitPos < 0) || ( entryPos >= noEntries ) || (entryPos < 0) ||
		( exitPos+numberOfExits-1 >= noExits ) || (exitPos+numberOfExits-1 < 0) || ( entryPos+numberOfEntries-1 >= noEntries ) || (entryPos+numberOfEntries-1 < 0))
	{
		printf("ERROR: in Connect()\n" );
		return false;
	}

	return true;

}

