////////////////////////////////////////////////////////////////////////////////
// File         : Hnet.h
// Subsystem    : Graph
// Component    : H-net
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : definition of the class Hnet
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 10.09.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
// 30.05.06    Josef Huebl       some extensions
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
//     
// The class Hnet implements isomophie classes of hierarchical s-nets.
// Hnet types a managed in a socalled Hnet universe.
// For detailed information see the document S-net.doc.
//     
////////////////////////////////////////////////////////////////////////////////


#if !defined(AFX_HNET_H__D113E337_D6C9_4F6D_9524_68554A4ED74D__INCLUDED_)
#define AFX_HNET_H__D113E337_D6C9_4F6D_9524_68554A4ED74D__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Snet.h"

typedef unsigned long IsomorphyID;  // changed later to unique ID 
typedef Node IsoTypeNode;

class HnetUniverse;

IsomorphyID GetNextIsomophyID( void );

class Hnet  
{
	friend class HnetUniverse;

public:
	Hnet( HnetUniverse *universe = NULL );
	Hnet( HnetUniverse *universe, IsoTypeNode &isoType );
	Hnet( int numOuterEntries, int numOuterExits, HnetUniverse *universe, IsoTypeNode &isoType, bool elementary = true );
	virtual ~Hnet();
	Hnet( const Hnet &hnet );
	Hnet& operator=( const Hnet &hnet);
	IsomorphyID GetIsomophyID( void ) const { return m_isoID; };
	HnetUniverse *GetHnetUniverse( void ) const { HnetUniverse *h = m_universe; 
	return m_universe; };

	// edit operations
	Node AddOuterEntry( void ) { CheckID(); return m_Snet.AddOuterEntry(); };
	Node AddOuterExit( void ) { CheckID(); return m_Snet.AddOuterExit(); };
	void DelOuterEntry( Node &y  ) { CheckID(); m_Snet.DelOuterEntry( y ); };
	void DelOuterExit( Node &x  ) { CheckID(); m_Snet.DelOuterExit( x ); };
	Node AddSubcentre( const IsoTypeNode &isoType );
	Node CopySubcentre( const Node &s );
	void DelSubcentre( Node &s  );
	void ChangeIsoType( const Node &s, const IsoTypeNode &isoType );
	Node Expand( Node &s );
	Node ExpandAll( void );
	void MakeFlat( void ) { while ( !Flat() ) ExpandAll(); };
	Edge Connect( const Node &x, const Node &y ) { CheckID(); return m_Snet.Connect( x, y); };
	void Disconnect( Edge &e ) { CheckID(); m_Snet.Disconnect( e ); }; 
	void SetElementary( void ) { if (!m_elementary) { CheckID(); m_elementary = true; } };	

	// inquiries
	IsoTypeNode GetIsoType( void ) const { return m_isoType[ m_Snet.Root() ]; };
	IsoTypeNode GetIsoType( const Node &s) const { return m_isoType[ s ]; };
	int NumberOfReferences(  const IsoTypeNode &isoType ) const;
	bool Elementary( void ) { return m_elementary; };
	bool Elementary( Node &s );
	bool Flat( void ); // true, if all subcenteres are elementary
	Snet &GetSnet( void ) { return m_Snet; };

	Node Root( void ) const { return m_Snet.Root(); };
	Node FirstSubcentre( void ) const { return m_Snet.FirstSubcentre(); }; 
	Node NextSubcentre( const Node &n) const { return m_Snet.NextSubcentre( n ); };
	Node LastSubcentre( void ) const { return m_Snet.LastSubcentre(); }; 
	Node PrevSubcentre( const Node &n) const { return m_Snet.PrevSubcentre( n ); }; 
	Node FirstEntry( Edge &e, const Node &s ) const  { return m_Snet.FirstEntry( e, s); }; // on call e is emty an edge
	Node NextEntry( Edge &e ) const { return m_Snet.NextEntry( e ); };
	Node FirstExit( Edge &e, const Node &s ) const  { return m_Snet.FirstExit( e, s); }; // on call e is emty an edge
	Node NextExit( Edge &e ) const { return m_Snet.NextExit( e ); };
	int NumberOfEntries( const Node &s ) const  { return m_Snet.NumberOfEntries( s ); };
	int NumberOfExits( const Node &s ) const { return m_Snet.NumberOfExits( s ); };
	Node FirstOuterEntry( Edge &e) const { return m_Snet.FirstOuterEntry( e ); }; // on call e is emty an edge
	Node NextOuterEntry( Edge &e ) const { return m_Snet.NextOuterEntry( e ); };
	Node FirstOuterExit( Edge &e ) const { return m_Snet.FirstOuterExit( e ); }; // on call e is emty an edge
	Node NextOuterExit( Edge &e ) const { return m_Snet.NextOuterExit( e ); };
	int NumberOfOuterEntries( void ) const { return m_Snet.NumberOfOuterEntries(); };
	int NumberOfOuterExits( void ) const { return m_Snet.NumberOfOuterExits(); };
	Edge FirstConnection( void ) const { return m_Snet.FirstConnection(); }; 
	Edge NextConnection( const Edge &e) const { return m_Snet.NextConnection( e ); };

private:
	Snet m_Snet;
	bool m_elementary;
	IsomorphyID m_isoID; // unique for each isomorphic s-net class
	bool m_virgin; // true if just copied

private:
	void CheckID( void );

protected:
	NodeAttribute< IsoTypeNode > m_isoType; // reference into the isomorphy type tree
	HnetUniverse *m_universe; // stores all isomorphy classes

protected:
	void Copy( const Hnet &hnet );
	NodeType GetNodeType( const Node &n ) const { return m_Snet.GetNodeType( n ); }


};

#endif // !defined(AFX_HNET_H__D113E337_D6C9_4F6D_9524_68554A4ED74D__INCLUDED_)
