////////////////////////////////////////////////////////////////////////////////
// File         : indexattr.cpp
// Subsystem    : Graph
// Component    : index based attributation
// Author       : Josef Huebl, (Josef.Huebl@sss.de)
// Description  : implementation of an index based attributation
////////////////////////////////////////////////////////////////////////////////
// Copyright (C) Triple-S GmbH (www.sss.de) 2003
// Alle Rechte vorbehalten. All Rights Reserved.
//
// The reproduction, transmission or use of this document or its contents is
// not permitted without express written authority.
// Offenders will be liable for damages. All rights, including rights created
// by patent grant or registration of a utility model or design, are reserved.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:
// 05.08.03    Josef Huebl       created
// 17.12.03    Josef Huebl       1. released version.
// 07.01.04    Josef Huebl	AttributeVector::AddAllAttributes() improvements
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Hints:      
//     
//     
////////////////////////////////////////////////////////////////////////////////


#include "indexattr.h"

// ****************************** IndexAdmin ************************************

Index IndexAdmin::NewIndex( void ) 
{	
	if (m_reusableIndexes.empty())
	{
		return m_nextUnusedIndex++;
	}
	else 
	{
		Index index = m_reusableIndexes.back();
		m_reusableIndexes.pop_back();
		return index;
	}
};
											
Index IndexAdmin::MaxIndex( void ) 
{	
	if ( m_nextUnusedIndex == 0 )
		return NONE_VALID_INDEX;
	else
		return (m_nextUnusedIndex - 1);
};
											
void IndexAdmin::DelIndex(Index index) 
{
	if ( index != NONE_VALID_INDEX )
	{
		//?? hier muesste eigentich geprft werden, on der index schon mal gelscht wurde.
		// eine Optimierung ist es, wenn vorausgesetzt wird, dass dies nie der Fall ist
		m_reusableIndexes.push_back( index ); 
	}
};	

void IndexAdmin::ClearIndexes( void ) 
{
	m_nextUnusedIndex = 0;
	m_reusableIndexes.clear(); 
};	

void IndexAdmin::ReInitialize(Index nextIndexInUse ) 
{
	m_nextUnusedIndex = nextIndexInUse;
	m_reusableIndexes.clear(); 
};	

// ****************************** AttributeIndexAdmin ************************************

AttributeID AttributeIndexAdmin::NewAttributeID( void )
{
	AttributeID id;

	id = m_attributeIDadmin.NewIndex();

	if ( m_attributeValueIndexAdmin.size() <= id )
		m_attributeValueIndexAdmin.resize( id+1, IndexAdmin() );  // really new ID
	else
		m_attributeValueIndexAdmin[id].ClearIndexes(); 

	return id;
};


AttributeID AttributeIndexAdmin::CopyAttributeID( AttributeID originalID )		
{
	AttributeID id;

	id = m_attributeIDadmin.NewIndex();

	if ( m_attributeValueIndexAdmin.size() <= id )
		m_attributeValueIndexAdmin.resize( id+1, IndexAdmin() );  // really new ID

	m_attributeValueIndexAdmin[ id ] = m_attributeValueIndexAdmin[ originalID ];

	return id;
};


AttributeID AttributeIndexAdmin::MaxAttributeID( void )
{
	Index id = m_attributeIDadmin.MaxIndex();

	if ( NONE_VALID_INDEX == id )
		return NONE_VALID_ATTRIBUTE_ID;
	else 
		return id;
};

void AttributeIndexAdmin::DelAttributeID( AttributeID id )
{
		m_attributeIDadmin.DelIndex( id );
		m_attributeValueIndexAdmin[ id ].ClearIndexes(); // prepare for reuse
};


AttributeValueIndex AttributeIndexAdmin::NewAttributeValueIndex( AttributeID id )
{
	return m_attributeValueIndexAdmin[ id ].NewIndex();
};


AttributeValueIndex AttributeIndexAdmin::MaxAttributeValueIndex( AttributeID id )
{
	return m_attributeValueIndexAdmin[ id ].MaxIndex();
};


void AttributeIndexAdmin::DelAttributeValueIndex( AttributeID id, AttributeValueIndex valueIndex )
{
	m_attributeValueIndexAdmin[ id ].DelIndex( valueIndex );
};

 
void AttributeIndexAdmin::ReInitializeAttributeValueIndexes( AttributeID id, AttributeValueIndex nextValueIndexInUse )
{
	m_attributeValueIndexAdmin[ id ].ReInitialize( nextValueIndexInUse );
};


void AttributeIndexAdmin::ClearAttributeValueIndexes( AttributeID id )
{
	m_attributeValueIndexAdmin[ id ].ClearIndexes();
};

void AttributeIndexAdmin::Clear( void )
{ 
	AttributeID maxID, id;

	maxID = m_attributeIDadmin.MaxIndex();

	if ( maxID != NONE_VALID_ATTRIBUTE_ID )
	{
		for( id=0; id<maxID; id++) 
			ClearAttributeValueIndexes( id );

		m_attributeIDadmin.ClearIndexes();
	}
};

// ****************************** AttributeVector ************************************

void AttributeVector::NewIndexedAttribute( AttributeID id ) 
{
	if (m_attributeVectorTablePtr != NULL )
	{
		IndexedAttribute x;
		AttributeIndexAdmin &attributeIndexAdmin = m_attributeVectorTablePtr->GetAttributeIndexAdmin();

		if ( m_indexedAttributes.size() <= id )
			m_indexedAttributes.resize( id+1, x );

		m_indexedAttributes[id].attributeValueIndex = attributeIndexAdmin.NewAttributeValueIndex( id );
		m_indexedAttributes[id].virgin = true;
	}
};


void AttributeVector::CopyIndexedAttribute( AttributeID idOriginal, AttributeID idCopy ) 
{
	if (m_attributeVectorTablePtr != NULL )
	{
		if ( m_indexedAttributes.size() <= idCopy )
		{
			IndexedAttribute x;
			m_indexedAttributes.resize( idCopy+1, x );
		}

		m_indexedAttributes[ idCopy ].attributeValueIndex = m_indexedAttributes[ idOriginal ].attributeValueIndex;
		m_indexedAttributes[ idCopy ].virgin = m_indexedAttributes[ idOriginal ].virgin;
	}
};


void AttributeVector::DelIndexedAttribute( AttributeID id ) 
{
	if (m_attributeVectorTablePtr != NULL )
	{
		AttributeIndexAdmin &attributeIndexAdmin = m_attributeVectorTablePtr->GetAttributeIndexAdmin();

		attributeIndexAdmin.DelAttributeValueIndex( id, m_indexedAttributes[id].attributeValueIndex );
		m_indexedAttributes[id].attributeValueIndex = NONE_VALID_ATTRIBUTE_VALUE_INDEX;
		m_indexedAttributes[id].virgin = true;
	}
};

void AttributeVector::AddAllAttributes( AttributeVectorTable *table ) 
{
	if (m_attributeVectorTablePtr == table ) // Attribute vector is already element of this table
		return;

	if (m_attributeVectorTablePtr != NULL ) // Attribute vector is element of another table
	{
		DelAllAttributes( );

		if (m_validPointerPtr != NULL )
		{
			*m_validPointerPtr = false;  // delete reference from element to this vector 
			m_validPointerPtr = NULL;  
		}
	}

	m_attributeVectorTablePtr = table;

	AttributeIndexAdmin &attributeIndexAdmin = table->GetAttributeIndexAdmin();
	AttributeID maxID = attributeIndexAdmin.MaxAttributeID();
	AttributeID id;
	IndexedAttribute x;

	if (maxID != NONE_VALID_ATTRIBUTE_ID )
	{
		if ( m_indexedAttributes.size() <= maxID )
			m_indexedAttributes.resize( maxID+1, x );

		for (id=0; id<=maxID; id++) // for every existing attribute ID
		{
			m_indexedAttributes[id].attributeValueIndex = attributeIndexAdmin.NewAttributeValueIndex( id );
			m_indexedAttributes[id].virgin = true;
		}
	}
};


void AttributeVector::DelAllAttributes( void ) 
{
	if (m_attributeVectorTablePtr == NULL ) // nothing to do
		return;

	AttributeIndexAdmin &attributeIndexAdmin = m_attributeVectorTablePtr->GetAttributeIndexAdmin();
	AttributeID id;
	AttributeID maxID = attributeIndexAdmin.MaxAttributeID();

	if (maxID != NONE_VALID_ATTRIBUTE_ID ) // at least one element?
	{
		for (id=0; id<=maxID; id++) // for every existing attribute ID
		{
			attributeIndexAdmin.DelAttributeValueIndex( id, m_indexedAttributes[ id ].attributeValueIndex ); 
		}
	}

	m_indexedAttributes.clear();
	m_attributeVectorTablePtr = NULL; 
};


// ****************************** AttributeVectorTable ************************************

AttributeVectorTable::~AttributeVectorTable()
{
	std::vector<AttributeVectorTable **>::iterator _it;

	// for all attributes
	for (_it=m_attributeVectorTablePtrPtr.begin(); _it!=m_attributeVectorTablePtrPtr.end(); ++_it)
	{
		if (*_it != NULL )
		{
			**(_it) = NULL; // delete reference from attribute to this table
			*(_it) = NULL; 
		}
	}
}


std::list<AttributeVector>::iterator AttributeVectorTable::NewAttributeVector( bool *validPointerPtr )
{
	AttributeVector vec( validPointerPtr );
	vec.AddAllAttributes( this ); // create index for every attribute
	std::list<AttributeVector>::iterator _it;
	m_attributeVectors.push_back( vec ); // Attention! The vector is copied here. This is not effective
	_it = m_attributeVectors.end();
	return --_it;
};


void AttributeVectorTable::DelAttributeVector( std::list<AttributeVector>::iterator elem)
{
	elem->DelAllAttributes(); // free indexes
	m_attributeVectors.erase( elem ); 
};



void AttributeVectorTable::DelAllAttributeVectors( void )
{
	m_attributeVectors.clear();
	m_attributeIndexAdmin.Clear();
};



AttributeID AttributeVectorTable::NewAttribute( AttributeVectorTable **attributeVectorTablePtrPtr )
{
	AttributeID id;
	Index i;
	std::list<AttributeVector>::iterator _it;

	id = m_attributeIndexAdmin.NewAttributeID();  // new indexed Id for attribute

	// store the attribute pointer and later set the reference to NULL in case of destruction
	if (id >= m_attributeVectorTablePtrPtr.size())
		m_attributeVectorTablePtrPtr.push_back( attributeVectorTablePtrPtr );
	else
		m_attributeVectorTablePtrPtr[ id ] = attributeVectorTablePtrPtr;

	// for all elements
	for (i=0, _it=m_attributeVectors.begin(); _it!=m_attributeVectors.end(); ++_it, ++i)
	{
		_it->NewIndexedAttribute( id ); 
	}

	m_attributeIndexAdmin.ReInitializeAttributeValueIndexes(id, i);

	return id;
};


AttributeID AttributeVectorTable::CopyAttribute( AttributeID originalID, AttributeVectorTable **copyAttributeVectorTablePtrPtr )
{
	std::list<AttributeVector>::iterator _it;
	Index i;
	AttributeID id = m_attributeIndexAdmin.CopyAttributeID( originalID );		

	// store the attribute pointer and later set the reference to NULL in case of destruction
	if (id >= m_attributeVectorTablePtrPtr.size())
		m_attributeVectorTablePtrPtr.push_back( copyAttributeVectorTablePtrPtr );
	else
		m_attributeVectorTablePtrPtr[ id ] = copyAttributeVectorTablePtrPtr;

	// for all elements
	for (i=0, _it=m_attributeVectors.begin(); _it!=m_attributeVectors.end(); ++_it, ++i)
	{
		_it->CopyIndexedAttribute( originalID, id ); 
	}

	return id;	
}


void AttributeVectorTable::DelAttribute( AttributeID id )
{
	std::list<AttributeVector>::iterator _it;

	//delete reference from attribute and the attribute pointer 
	if (m_attributeVectorTablePtrPtr[ id ] != NULL )
	{
		*(m_attributeVectorTablePtrPtr[ id ]) = NULL;
		m_attributeVectorTablePtrPtr[ id ] = NULL;
	}


	// for all elements
	for (_it=m_attributeVectors.begin(); _it!=m_attributeVectors.end(); ++_it)
	{
		_it->DelIndexedAttribute( id ); 
	}

	m_attributeIndexAdmin.ClearAttributeValueIndexes( id );
};


void AttributeVectorTable::ReinitialiseAttribute( AttributeID id )
{
	std::list<AttributeVector>::iterator _it;

	for (_it=m_attributeVectors.begin(); _it!=m_attributeVectors.end(); ++_it)
	{
		_it->ResetIndexedAttribute( id ); 
	}
};

// ****************************** AttributedElement ************************************

void AttributedElement::NewAttributeVector( AttributeVectorTable *ptr)
{
	m_validPointer = true;
	m_attributeVectorPtr = ptr->NewAttributeVector( &m_validPointer );
}


void AttributedElement::DelAttributeVector( void )
{
	if (m_validPointer)
	{
		m_validPointer = false; 
		m_attributeVectorPtr->GetAttributeVectorTablePtr()->DelAttributeVector( m_attributeVectorPtr );;
	}
};


// ****************************** AttributeBase ************************************

AttributeValueIndex AttributeBase::CheckAttributeValue( const std::list<AttributeVector>::iterator vectorPtr ) const
{
	if ( (m_attributeVectorTablePtr != NULL) && (m_attributeVectorTablePtr != vectorPtr->GetAttributeVectorTablePtr()))
	{
//		AttributeVectorTable *ptr = vectorPtr->GetAttributeVectorTablePtr(); // for debugging only
		m_attributeVectorTablePtr->DelAttribute( m_attributID ); // attribute belongs to another table
		m_attributeVectorTablePtr = NULL;
	}

	if (m_attributeVectorTablePtr == NULL) // value set first time 
	{
		m_attributeVectorTablePtr = vectorPtr->GetAttributeVectorTablePtr();	// store pointer to table for destructor 
		m_attributID = m_attributeVectorTablePtr->NewAttribute( &m_attributeVectorTablePtr ); // new attribute
	}

	AttributeValueIndex max = m_attributeVectorTablePtr->MaxAttributeValueIndex( m_attributID ); 
	ResizeUpTo( max );	// this is absolutely neccessary; otherwise the statement attr[ x ] = attr[ y ]; can fail,
						// because a reallocation can destroy the memory

	IndexedAttribute indexedAttribute = vectorPtr->GetIndexedAttribute( m_attributID );
	AttributeValueIndex valueIndex = indexedAttribute.attributeValueIndex;

	if ( indexedAttribute.virgin ) // set default value for virgin indexes 
	{								
		vectorPtr->UseIndexedAttribute( m_attributID ); // set virgin = false;
		ResetToDefaultValue( valueIndex );
	}

//	indexedAttribute = vectorPtr->GetIndexedAttribute( m_attributID ); // for test

	return valueIndex;
};


IndexedAttribute AttributeBase::GetIndexedAttribute( const std::list<AttributeVector>::iterator vectorPtr ) const
{
	if ( (m_attributeVectorTablePtr != NULL) && (m_attributeVectorTablePtr != vectorPtr->GetAttributeVectorTablePtr()))
	{
		m_attributeVectorTablePtr->DelAttribute( m_attributID ); // attribute belongs to another table
		m_attributeVectorTablePtr = NULL;
	}

	if (m_attributeVectorTablePtr == NULL) // value set first time 
	{
		m_attributeVectorTablePtr = vectorPtr->GetAttributeVectorTablePtr();	// store pointer to table for destructor 
		m_attributID = m_attributeVectorTablePtr->NewAttribute( &m_attributeVectorTablePtr ); // new attribute
	}

	return vectorPtr->GetIndexedAttribute( m_attributID );
};


